"""
Hyperbolic Householder transformations for Up- and Downdating Cholesky factorizations.
"""

__version__ = "1.0.2a3"

from ._hyhound import *  # noqa: F403
from ._hyhound import __version__ as __c_version__  # noqa: F401
